@@ -3,7 +3,7 @@
 
  Part of the Routino routing software.
  ******************/ /******************
- This file Copyright 2009-2012 Andrew M. Bishop
+ This file Copyright 2009-2015, 2017, 2019, 2023 Andrew M. Bishop
 
  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU Affero General Public License as published by
@@ -23,7 +23,6 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
-#include <assert.h>
 
 #if defined(USE_PTHREADS) && USE_PTHREADS
 #include <pthread.h>
@@ -31,6 +30,7 @@
 
 #include "types.h"
 
+#include "logging.h"
 #include "files.h"
 #include "sorting.h"
 
@@ -52,18 +52,22 @@ extern int option_filesort_threads;
 /*+ A data type for holding data for a thread. +*/
 typedef struct _thread_data
  {
+#if defined(USE_PTHREADS) && USE_PTHREADS
+
   pthread_t thread;             /*+ The thread identifier. +*/
 
   int       running;            /*+ A flag indicating the current state of the thread. +*/
 
-  void     *data;               /*+ The main data array. +*/
+#endif
+
+  char     *data;               /*+ The main data array. +*/
   void    **datap;              /*+ An array of pointers to the data objects. +*/
   size_t    n;                  /*+ The number of pointers. +*/
 
-  char    *filename;            /*+ The name of the file to write the results to. +*/
+  int       fd;                 /*+ The file descriptor of the file to write the results to. +*/
 
-  size_t   itemsize;            /*+ The size of each item. +*/
-  int    (*compare)(const void*,const void*); /*+ The comparison function. +*/
+  size_t    itemsize;           /*+ The size of each item. +*/
+  int     (*compare)(const void*,const void*); /*+ The comparison function. +*/
  }
  thread_data;
 
@@ -72,7 +76,7 @@ typedef struct _thread_data
 #if defined(USE_PTHREADS) && USE_PTHREADS
 
 static pthread_mutex_t running_mutex = PTHREAD_MUTEX_INITIALIZER;
-static pthread_cond_t running_cond = PTHREAD_COND_INITIALIZER;
+static pthread_cond_t  running_cond  = PTHREAD_COND_INITIALIZER;
 
 #endif
 
@@ -100,41 +104,65 @@ static void *filesort_vary_heapsort_thread(thread_data *thread);
 
   size_t itemsize The size of each item in the file that needs sorting.
 
-  int (*compare)(const void*, const void*) The comparison function (identical to qsort if the
-                                           data to be sorted is an array of things not pointers).
+  int (*pre_sort_function)(void *,index_t) If non-NULL then this function is called for
+     each item before they have been sorted.  The second parameter is the number of objects
+     previously read from the input file.  If the function returns 1 then the object is kept
+     and it is sorted, otherwise it is ignored.
 
-  int (*keep)(void *,index_t) If non-NULL then this function is called for each item, if it
-                              returns 1 then the object is kept and written to the output file.
+  int (*compare_function)(const void*, const void*) The comparison function.  This is identical
+     to qsort if the data to be sorted is an array of things not pointers.
+
+  int (*post_sort_function)(void *,index_t) If non-NULL then this function is called for
+     each item after they have been sorted.  The second parameter is the number of objects
+     already written to the output file.  If the function returns 1 then the object is written
+     to the output file., otherwise it is ignored.
   ++++++++++++++++++++++++++++++++++++++*/
 
-index_t filesort_fixed(int fd_in,int fd_out,size_t itemsize,int (*compare)(const void*,const void*),int (*keep)(void*,index_t))
+index_t filesort_fixed(int fd_in,int fd_out,size_t itemsize,int (*pre_sort_function)(void*,index_t),
+                                                            int (*compare_function)(const void*,const void*),
+                                                            int (*post_sort_function)(void*,index_t))
 {
  int *fds=NULL,*heap=NULL;
  int nfiles=0,ndata=0;
- index_t count=0,total=0;
- size_t nitems=option_filesort_ramsize/(option_filesort_threads*(itemsize+sizeof(void*)));
- void *data,**datap;
+ index_t count_out=0,count_in=0,total=0;
+ size_t nitems,item;
+ char *data;
+ void **datap;
  thread_data *threads;
  int i,more=1;
+ char *filename=(char*)malloc_logassert(strlen(option_tmpdirname)+24);
 #if defined(USE_PTHREADS) && USE_PTHREADS
  int nthreads=0;
 #endif
 
  /* Allocate the RAM buffer and other bits */
 
- threads=(thread_data*)malloc(option_filesort_threads*sizeof(thread_data));
+ nitems=(size_t)SizeFileFD(fd_in)/itemsize;
+
+ if(nitems==0)
+    return(0);
+
+ if((nitems*(itemsize+sizeof(void*)))<(option_filesort_ramsize/option_filesort_threads))
+    /* use one thread */;
+ else if((nitems*(itemsize+sizeof(void*)))<option_filesort_ramsize)
+    nitems=1+nitems/option_filesort_threads;
+ else
+    nitems=option_filesort_ramsize/(option_filesort_threads*(itemsize+sizeof(void*)));
+
+ threads=(thread_data*)calloc_logassert(option_filesort_threads,sizeof(thread_data));
 
  for(i=0;i<option_filesort_threads;i++)
    {
-    threads[i].running=0;
+    threads[i].fd=-1;
 
-    threads[i].data=malloc(nitems*itemsize);
-    threads[i].datap=malloc(nitems*sizeof(void*));
+    threads[i].data=malloc_logassert(nitems*itemsize);
+    threads[i].datap=malloc_logassert(nitems*sizeof(void*));
 
-    threads[i].filename=(char*)malloc(strlen(option_tmpdirname)+24);
+    log_malloc(threads[i].data ,nitems*itemsize);
+    log_malloc(threads[i].datap,nitems*sizeof(void*));
 
     threads[i].itemsize=itemsize;
-    threads[i].compare=compare;
+    threads[i].compare=compare_function;
    }
 
  /* Loop around, fill the buffer, sort the data and write a temporary file */
@@ -147,7 +175,31 @@ index_t filesort_fixed(int fd_in,int fd_out,size_t itemsize,int (*compare)(const
 
     if(option_filesort_threads>1)
       {
-       /* Find a spare slot (one *must* be unused at all times) */
+       /* If all threads are in use wait for an existing thread to finish */
+
+       if(nthreads==option_filesort_threads)
+         {
+          pthread_mutex_lock(&running_mutex);
+
+          while(nthreads==option_filesort_threads)
+            {
+             for(i=0;i<option_filesort_threads;i++)
+                if(threads[i].running==2)
+                  {
+                   pthread_join(threads[i].thread,NULL);
+                   threads[i].running=0;
+                   CloseFileBuffered(threads[i].fd);
+                   nthreads--;
+                  }
+
+             if(nthreads==option_filesort_threads)
+                pthread_cond_wait(&running_cond,&running_mutex);
+            }
+
+          pthread_mutex_unlock(&running_mutex);
+         }
+
+       /* Find a spare slot */
 
        pthread_mutex_lock(&running_mutex);
 
@@ -162,20 +214,26 @@ index_t filesort_fixed(int fd_in,int fd_out,size_t itemsize,int (*compare)(const
 
     /* Read in the data and create pointers */
 
-    for(i=0;i<nitems;i++)
+    for(item=0;item<nitems;)
       {
-       threads[thread].datap[i]=threads[thread].data+i*itemsize;
+       threads[thread].datap[item]=threads[thread].data+item*itemsize;
 
-       if(ReadFile(fd_in,threads[thread].datap[i],itemsize))
+       if(ReadFileBuffered(fd_in,threads[thread].datap[item],itemsize))
          {
           more=0;
           break;
          }
 
-       total++;
+       if(!pre_sort_function || pre_sort_function(threads[thread].datap[item],count_in))
+         {
+          item++;
+          total++;
+         }
+
+       count_in++;
       }
 
-    threads[thread].n=i;
+    threads[thread].n=item;
 
     /* Shortcut if there is no previous data and no more data (i.e. no data at all) */
 
@@ -187,56 +245,45 @@ index_t filesort_fixed(int fd_in,int fd_out,size_t itemsize,int (*compare)(const
     if(threads[thread].n==0)
        break;
 
-    /* Sort the data pointers using a heap sort (potentially in a thread) */
+    /* Shortcut if only one file, don't write to disk */
 
-    sprintf(threads[thread].filename,"%s/filesort.%d.tmp",option_tmpdirname,nfiles);
+    if(more==0 && nfiles==0)
+       filesort_heapsort(threads[thread].datap,threads[thread].n,threads[thread].compare);
+
+    else
+      {
+       /* Create the file descriptor (not thread-safe) */
+
+       sprintf(filename,"%s/filesort.%d.tmp",option_tmpdirname,nfiles);
+
+       threads[thread].fd=OpenFileBufferedNew(filename);
+
+       if(option_filesort_threads==1)
+         {
+          filesort_fixed_heapsort_thread(&threads[thread]);
+
+          CloseFileBuffered(threads[thread].fd);
+         }
 
 #if defined(USE_PTHREADS) && USE_PTHREADS
 
-    /* Shortcut if only one file, don't write to disk */
-
-    if(more==0 && nfiles==0)
-       filesort_heapsort(threads[thread].datap,threads[thread].n,threads[thread].compare);
-    else if(option_filesort_threads>1)
-      {
-       pthread_mutex_lock(&running_mutex);
-
-       while(nthreads==(option_filesort_threads-1))
+       else
          {
-          for(i=0;i<option_filesort_threads;i++)
-             if(threads[i].running==2)
-               {
-                pthread_join(threads[i].thread,NULL);
-                threads[i].running=0;
-                nthreads--;
-               }
+          pthread_mutex_lock(&running_mutex);
 
-          if(nthreads==(option_filesort_threads-1))
-             pthread_cond_wait(&running_cond,&running_mutex);
+          threads[thread].running=1;
+
+          pthread_mutex_unlock(&running_mutex);
+
+          pthread_create(&threads[thread].thread,NULL,(void* (*)(void*))filesort_fixed_heapsort_thread,&threads[thread]);
+
+          nthreads++;
          }
 
-       threads[thread].running=1;
-
-       pthread_mutex_unlock(&running_mutex);
-
-       pthread_create(&threads[thread].thread,NULL,(void* (*)(void*))filesort_fixed_heapsort_thread,&threads[thread]);
-
-       nthreads++;
-      }
-    else
-       filesort_fixed_heapsort_thread(&threads[thread]);
-
-#else
-
-    /* Shortcut if only one file, don't write to disk */
-
-    if(more==0 && nfiles==0)
-       filesort_heapsort(threads[thread].datap,threads[thread].n,threads[thread].compare);
-    else
-       filesort_fixed_heapsort_thread(&threads[thread]);
-
 #endif
 
+      }
+
     nfiles++;
    }
  while(more);
@@ -256,6 +303,7 @@ index_t filesort_fixed(int fd_in,int fd_out,size_t itemsize,int (*compare)(const
          {
           pthread_join(threads[i].thread,NULL);
           threads[i].running=0;
+          CloseFileBuffered(threads[i].fd);
           nthreads--;
          }
 
@@ -264,46 +312,49 @@ index_t filesort_fixed(int fd_in,int fd_out,size_t itemsize,int (*compare)(const
 
 #endif
 
+ /* Shortcut if there are no files */
+
+ if(nfiles==0)
+    goto tidy_and_exit;
+
  /* Shortcut if only one file, lucky for us we still have the data in RAM) */
 
  if(nfiles==1)
    {
-    for(i=0;i<threads[0].n;i++)
+    for(item=0;item<threads[0].n;item++)
       {
-       if(!keep || keep(threads[0].datap[i],count))
+       if(!post_sort_function || post_sort_function(threads[0].datap[item],count_out))
          {
-          WriteFile(fd_out,threads[0].datap[i],itemsize);
-          count++;
+          WriteFileBuffered(fd_out,threads[0].datap[item],itemsize);
+          count_out++;
          }
       }
 
-    DeleteFile(threads[0].filename);
+    DeleteFile(filename);
 
     goto tidy_and_exit;
    }
 
  /* Check that number of files is less than file size */
 
- assert(nfiles<nitems);
+ logassert((unsigned)nfiles<nitems,"Too many temporary files (use more sorting memory?)");
 
  /* Open all of the temporary files */
 
- fds=(int*)malloc(nfiles*sizeof(int));
+ fds=(int*)malloc_logassert(nfiles*sizeof(int));
 
  for(i=0;i<nfiles;i++)
    {
-    char *filename=threads[0].filename;
-
     sprintf(filename,"%s/filesort.%d.tmp",option_tmpdirname,i);
 
-    fds[i]=ReOpenFile(filename);
+    fds[i]=ReOpenFileBuffered(filename);
 
     DeleteFile(filename);
    }
 
  /* Perform an n-way merge using a binary heap */
 
- heap=(int*)malloc((1+nfiles)*sizeof(int));
+ heap=(int*)malloc_logassert((1+nfiles)*sizeof(int));
 
  data =threads[0].data;
  datap=threads[0].datap;
@@ -316,7 +367,7 @@ index_t filesort_fixed(int fd_in,int fd_out,size_t itemsize,int (*compare)(const
 
     datap[i]=data+i*itemsize;
 
-    ReadFile(fds[i],datap[i],itemsize);
+    ReadFileBuffered(fds[i],datap[i],itemsize);
 
     index=i+1;
 
@@ -331,7 +382,7 @@ index_t filesort_fixed(int fd_in,int fd_out,size_t itemsize,int (*compare)(const
 
        newindex=index/2;
 
-       if(compare(datap[heap[index]],datap[heap[newindex]])>=0)
+       if(compare_function(datap[heap[index]],datap[heap[newindex]])>=0)
           break;
 
        temp=heap[index];
@@ -350,13 +401,13 @@ index_t filesort_fixed(int fd_in,int fd_out,size_t itemsize,int (*compare)(const
    {
     int index=1;
 
-    if(!keep || keep(datap[heap[index]],count))
+    if(!post_sort_function || post_sort_function(datap[heap[index]],count_out))
       {
-       WriteFile(fd_out,datap[heap[index]],itemsize);
-       count++;
+       WriteFileBuffered(fd_out,datap[heap[index]],itemsize);
+       count_out++;
       }
 
-    if(ReadFile(fds[heap[index]],datap[heap[index]],itemsize))
+    if(ReadFileBuffered(fds[heap[index]],datap[heap[index]],itemsize))
       {
        heap[index]=heap[ndata];
        ndata--;
@@ -371,10 +422,10 @@ index_t filesort_fixed(int fd_in,int fd_out,size_t itemsize,int (*compare)(const
 
        newindex=2*index;
 
-       if(compare(datap[heap[newindex]],datap[heap[newindex+1]])>=0)
+       if(compare_function(datap[heap[newindex]],datap[heap[newindex+1]])>=0)
           newindex=newindex+1;
 
-       if(compare(datap[heap[index]],datap[heap[newindex]])<=0)
+       if(compare_function(datap[heap[index]],datap[heap[newindex]])<=0)
           break;
 
        temp=heap[newindex];
@@ -391,7 +442,7 @@ index_t filesort_fixed(int fd_in,int fd_out,size_t itemsize,int (*compare)(const
 
        newindex=2*index;
 
-       if(compare(datap[heap[index]],datap[heap[newindex]])<=0)
+       if(compare_function(datap[heap[index]],datap[heap[newindex]])<=0)
           ; /* break */
        else
          {
@@ -410,7 +461,7 @@ index_t filesort_fixed(int fd_in,int fd_out,size_t itemsize,int (*compare)(const
  if(fds)
    {
     for(i=0;i<nfiles;i++)
-       CloseFile(fds[i]);
+       CloseFileBuffered(fds[i]);
     free(fds);
    }
 
@@ -419,13 +470,18 @@ index_t filesort_fixed(int fd_in,int fd_out,size_t itemsize,int (*compare)(const
 
  for(i=0;i<option_filesort_threads;i++)
    {
+    log_free(threads[i].data);
+    log_free(threads[i].datap);
+
     free(threads[i].data);
     free(threads[i].datap);
-
-    free(threads[i].filename);
    }
 
- return(count);
+ free(threads);
+
+ free(filename);
+
+ return(count_out);
 }
 
 
@@ -445,46 +501,75 @@ index_t filesort_fixed(int fd_in,int fd_out,size_t itemsize,int (*compare)(const
 
   int fd_out The file descriptor of the output file (opened for writing and empty).
 
-  int (*compare)(const void*, const void*) The comparison function (identical to qsort if the
-                                           data to be sorted is an array of things not pointers).
+  int (*pre_sort_function)(void *,index_t) If non-NULL then this function is called for
+     each item before they have been sorted.  The second parameter is the number of objects
+     previously read from the input file.  If the function returns 1 then the object is kept
+     and it is sorted, otherwise it is ignored.
 
-  int (*keep)(void *,index_t) If non-NULL then this function is called for each item, if it
-                              returns 1 then the object is kept and written to the output file.
+  int (*compare_function)(const void*, const void*) The comparison function.  This is identical
+     to qsort if the data to be sorted is an array of things not pointers.
+
+  int (*post_sort_function)(void *,index_t) If non-NULL then this function is called for
+     each item after they have been sorted.  The second parameter is the number of objects
+     already written to the output file.  If the function returns 1 then the object is written
+     to the output file., otherwise it is ignored.
   ++++++++++++++++++++++++++++++++++++++*/
 
-index_t filesort_vary(int fd_in,int fd_out,int (*compare)(const void*,const void*),int (*keep)(void*,index_t))
+index_t filesort_vary(int fd_in,int fd_out,int (*pre_sort_function)(void*,index_t),
+                                           int (*compare_function)(const void*,const void*),
+                                           int (*post_sort_function)(void*,index_t))
 {
  int *fds=NULL,*heap=NULL;
  int nfiles=0,ndata=0;
- index_t count=0,total=0;
- size_t datasize=option_filesort_ramsize/option_filesort_threads;
+ index_t count_out=0,count_in=0,total=0;
+ size_t datasize,item;
  FILESORT_VARINT nextitemsize,largestitemsize=0;
- void *data,**datap;
+ char *data;
+ void **datap;
  thread_data *threads;
  int i,more=1;
+ char *filename=(char*)malloc_logassert(strlen(option_tmpdirname)+24);
 #if defined(USE_PTHREADS) && USE_PTHREADS
  int nthreads=0;
 #endif
 
  /* Allocate the RAM buffer and other bits */
 
- threads=(thread_data*)malloc(option_filesort_threads*sizeof(thread_data));
+ datasize=(size_t)SizeFileFD(fd_in);
+
+ if(datasize==0)
+    return(0);
+
+ /* We can not know in advance how many data items there are.  Each
+    one will require RAM for data, FILESORT_VARALIGN and sizeof(void*)
+    Assume that data+FILESORT_VARALIGN+sizeof(void*) is 4*data. */
+
+ if((datasize*4)<option_filesort_ramsize/option_filesort_threads)
+    /* use one thread */;
+ else if((datasize*4)<option_filesort_ramsize)
+    datasize=(datasize*4)/option_filesort_threads;
+ else
+    datasize=option_filesort_ramsize/option_filesort_threads;
+
+ datasize=FILESORT_VARALIGN*((datasize+FILESORT_VARALIGN-1)/FILESORT_VARALIGN);
+
+ threads=(thread_data*)calloc_logassert(option_filesort_threads,sizeof(thread_data));
 
  for(i=0;i<option_filesort_threads;i++)
    {
-    threads[i].running=0;
+    threads[i].fd=-1;
 
-    threads[i].data=malloc(datasize);
+    threads[i].data=malloc_logassert(datasize);
     threads[i].datap=NULL;
 
-    threads[i].filename=(char*)malloc(strlen(option_tmpdirname)+24);
+    log_malloc(threads[i].data,datasize);
 
-    threads[i].compare=compare;
+    threads[i].compare=compare_function;
    }
 
  /* Loop around, fill the buffer, sort the data and write a temporary file */
 
- if(ReadFile(fd_in,&nextitemsize,FILESORT_VARSIZE))    /* Always have the next item size known in advance */
+ if(ReadFileBuffered(fd_in,&nextitemsize,FILESORT_VARSIZE))    /* Always have the next item size known in advance */
     goto tidy_and_exit;
 
  do
@@ -496,7 +581,31 @@ index_t filesort_vary(int fd_in,int fd_out,int (*compare)(const void*,const void
 
     if(option_filesort_threads>1)
       {
-       /* Find a spare slot (one *must* be unused at all times) */
+       /* If all threads are in use wait for an existing thread to finish */
+
+       if(nthreads==option_filesort_threads)
+         {
+          pthread_mutex_lock(&running_mutex);
+
+          while(nthreads==option_filesort_threads)
+            {
+             for(i=0;i<option_filesort_threads;i++)
+                if(threads[i].running==2)
+                  {
+                   pthread_join(threads[i].thread,NULL);
+                   threads[i].running=0;
+                   CloseFileBuffered(threads[i].fd);
+                   nthreads--;
+                  }
+
+             if(nthreads==option_filesort_threads)
+                pthread_cond_wait(&running_cond,&running_mutex);
+            }
+
+          pthread_mutex_unlock(&running_mutex);
+         }
+
+       /* Find a spare slot */
 
        pthread_mutex_lock(&running_mutex);
 
@@ -509,36 +618,43 @@ index_t filesort_vary(int fd_in,int fd_out,int (*compare)(const void*,const void
 
 #endif
 
-    threads[thread].datap=threads[thread].data+datasize;
+    threads[thread].datap=(void**)(threads[thread].data+datasize);
 
     threads[thread].n=0;
 
     /* Read in the data and create pointers */
 
-    while((ramused+FILESORT_VARSIZE+nextitemsize)<=((void*)threads[thread].datap-sizeof(void*)-threads[thread].data))
+    while((ramused+FILESORT_VARSIZE+nextitemsize)<=(size_t)((char*)threads[thread].datap-sizeof(void*)-threads[thread].data))
       {
        FILESORT_VARINT itemsize=nextitemsize;
 
-       if(itemsize>largestitemsize)
-          largestitemsize=itemsize;
-
        *(FILESORT_VARINT*)(threads[thread].data+ramused)=itemsize;
 
        ramused+=FILESORT_VARSIZE;
 
-       ReadFile(fd_in,threads[thread].data+ramused,itemsize);
+       ReadFileBuffered(fd_in,threads[thread].data+ramused,itemsize);
 
-       *--threads[thread].datap=threads[thread].data+ramused; /* points to real data */
+       if(!pre_sort_function || pre_sort_function(threads[thread].data+ramused,count_in))
+         {
+          *--threads[thread].datap=threads[thread].data+ramused; /* points to real data */
 
-       ramused+=itemsize;
+          if(itemsize>largestitemsize)
+             largestitemsize=itemsize;
 
-       ramused =FILESORT_VARALIGN*((ramused+FILESORT_VARSIZE-1)/FILESORT_VARALIGN);
-       ramused+=FILESORT_VARALIGN-FILESORT_VARSIZE;
+          ramused+=itemsize;
 
-       total++;
-       threads[thread].n++;
+          ramused =FILESORT_VARALIGN*((ramused+FILESORT_VARALIGN-1)/FILESORT_VARALIGN);
+          ramused+=FILESORT_VARALIGN-FILESORT_VARSIZE;
 
-       if(ReadFile(fd_in,&nextitemsize,FILESORT_VARSIZE))
+          total++;
+          threads[thread].n++;
+         }
+       else
+          ramused-=FILESORT_VARSIZE;
+
+       count_in++;
+
+       if(ReadFileBuffered(fd_in,&nextitemsize,FILESORT_VARSIZE))
          {
           more=0;
           break;
@@ -550,59 +666,45 @@ index_t filesort_vary(int fd_in,int fd_out,int (*compare)(const void*,const void
     if(threads[thread].n==0)
        break;
 
-    /* Sort the data pointers using a heap sort (potentially in a thread) */
+    /* Shortcut if only one file, don't write to disk */
 
     if(more==0 && nfiles==0)
-       threads[thread].filename[0]=0;
+       filesort_heapsort(threads[thread].datap,threads[thread].n,threads[thread].compare);
+
     else
-       sprintf(threads[thread].filename,"%s/filesort.%d.tmp",option_tmpdirname,nfiles);
+      {
+       /* Create the file descriptor (not thread-safe) */
+
+       sprintf(filename,"%s/filesort.%d.tmp",option_tmpdirname,nfiles);
+
+       threads[thread].fd=OpenFileBufferedNew(filename);
+
+       if(option_filesort_threads==1)
+         {
+          filesort_vary_heapsort_thread(&threads[thread]);
+
+          CloseFileBuffered(threads[thread].fd);
+         }
 
 #if defined(USE_PTHREADS) && USE_PTHREADS
 
-    /* Shortcut if only one file, don't write to disk */
-
-    if(more==0 && nfiles==0)
-       filesort_heapsort(threads[thread].datap,threads[thread].n,threads[thread].compare);
-    else if(option_filesort_threads>1)
-      {
-       pthread_mutex_lock(&running_mutex);
-
-       while(nthreads==(option_filesort_threads-1))
+       else
          {
-          for(i=0;i<option_filesort_threads;i++)
-             if(threads[i].running==2)
-               {
-                pthread_join(threads[i].thread,NULL);
-                threads[i].running=0;
-                nthreads--;
-               }
+          pthread_mutex_lock(&running_mutex);
 
-          if(nthreads==(option_filesort_threads-1))
-             pthread_cond_wait(&running_cond,&running_mutex);
+          threads[thread].running=1;
+
+          pthread_mutex_unlock(&running_mutex);
+
+          pthread_create(&threads[thread].thread,NULL,(void* (*)(void*))filesort_vary_heapsort_thread,&threads[thread]);
+
+          nthreads++;
          }
 
-       threads[thread].running=1;
-
-       pthread_mutex_unlock(&running_mutex);
-
-       pthread_create(&threads[thread].thread,NULL,(void* (*)(void*))filesort_vary_heapsort_thread,&threads[thread]);
-
-       nthreads++;
-      }
-    else
-       filesort_vary_heapsort_thread(&threads[thread]);
-
-#else
-
-    /* Shortcut if only one file, don't write to disk */
-
-    if(more==0 && nfiles==0)
-       filesort_heapsort(threads[thread].datap,threads[thread].n,threads[thread].compare);
-    else
-       filesort_vary_heapsort_thread(&threads[thread]);
-
 #endif
 
+      }
+
     nfiles++;
    }
  while(more);
@@ -622,6 +724,7 @@ index_t filesort_vary(int fd_in,int fd_out,int (*compare)(const void*,const void
          {
           pthread_join(threads[i].thread,NULL);
           threads[i].running=0;
+          CloseFileBuffered(threads[i].fd);
           nthreads--;
          }
 
@@ -630,53 +733,56 @@ index_t filesort_vary(int fd_in,int fd_out,int (*compare)(const void*,const void
 
 #endif
 
+ /* Shortcut if there are no files */
+
+ if(nfiles==0)
+    goto tidy_and_exit;
+
  /* Shortcut if only one file, lucky for us we still have the data in RAM) */
 
  if(nfiles==1)
    {
-    for(i=0;i<threads[0].n;i++)
+    for(item=0;item<threads[0].n;item++)
       {
-       if(!keep || keep(threads[0].datap[i],count))
+       if(!post_sort_function || post_sort_function(threads[0].datap[item],count_out))
          {
-          FILESORT_VARINT itemsize=*(FILESORT_VARINT*)(threads[0].datap[i]-FILESORT_VARSIZE);
+          FILESORT_VARINT itemsize=*(FILESORT_VARINT*)((char*)threads[0].datap[item]-FILESORT_VARSIZE);
 
-          WriteFile(fd_out,threads[0].datap[i]-FILESORT_VARSIZE,itemsize+FILESORT_VARSIZE);
-          count++;
+          WriteFileBuffered(fd_out,(char*)threads[0].datap[item]-FILESORT_VARSIZE,itemsize+FILESORT_VARSIZE);
+          count_out++;
          }
       }
 
-    DeleteFile(threads[0].filename);
+    DeleteFile(filename);
 
     goto tidy_and_exit;
    }
 
  /* Check that number of files is less than file size */
 
- largestitemsize=FILESORT_VARALIGN*(1+(largestitemsize+FILESORT_VARALIGN-FILESORT_VARSIZE)/FILESORT_VARALIGN);
+ largestitemsize=FILESORT_VARALIGN*((largestitemsize+FILESORT_VARALIGN-1)/FILESORT_VARALIGN);
 
- assert(nfiles<((datasize-nfiles*sizeof(void*))/largestitemsize));
+ logassert((unsigned)nfiles<((datasize-nfiles*sizeof(void*))/(FILESORT_VARALIGN+largestitemsize)),"Too many temporary files (use more sorting memory?)");
 
  /* Open all of the temporary files */
 
- fds=(int*)malloc(nfiles*sizeof(int));
+ fds=(int*)malloc_logassert(nfiles*sizeof(int));
 
  for(i=0;i<nfiles;i++)
    {
-    char *filename=threads[0].filename;
-
     sprintf(filename,"%s/filesort.%d.tmp",option_tmpdirname,i);
 
-    fds[i]=ReOpenFile(filename);
+    fds[i]=ReOpenFileBuffered(filename);
 
     DeleteFile(filename);
    }
 
  /* Perform an n-way merge using a binary heap */
 
- heap=(int*)malloc((1+nfiles)*sizeof(int));
+ heap=(int*)malloc_logassert((1+nfiles)*sizeof(int));
 
  data=threads[0].data;
- datap=data+datasize-nfiles*sizeof(void*);
+ datap=(void**)(data+datasize-nfiles*sizeof(void*));
 
  /* Fill the heap to start with */
 
@@ -685,13 +791,13 @@ index_t filesort_vary(int fd_in,int fd_out,int (*compare)(const void*,const void
     int index;
     FILESORT_VARINT itemsize;
 
-    datap[i]=data+FILESORT_VARALIGN-FILESORT_VARSIZE+i*largestitemsize;
+    datap[i]=data+FILESORT_VARALIGN+i*(largestitemsize+FILESORT_VARALIGN);
 
-    ReadFile(fds[i],&itemsize,FILESORT_VARSIZE);
+    ReadFileBuffered(fds[i],&itemsize,FILESORT_VARSIZE);
 
-    *(FILESORT_VARINT*)(datap[i]-FILESORT_VARSIZE)=itemsize;
+    *(FILESORT_VARINT*)((char*)datap[i]-FILESORT_VARSIZE)=itemsize;
 
-    ReadFile(fds[i],datap[i],itemsize);
+    ReadFileBuffered(fds[i],datap[i],itemsize);
 
     index=i+1;
 
@@ -706,7 +812,7 @@ index_t filesort_vary(int fd_in,int fd_out,int (*compare)(const void*,const void
 
        newindex=index/2;
 
-       if(compare(datap[heap[index]],datap[heap[newindex]])>=0)
+       if(compare_function(datap[heap[index]],datap[heap[newindex]])>=0)
           break;
 
        temp=heap[index];
@@ -726,24 +832,24 @@ index_t filesort_vary(int fd_in,int fd_out,int (*compare)(const void*,const void
     int index=1;
     FILESORT_VARINT itemsize;
 
-    if(!keep || keep(datap[heap[index]],count))
+    if(!post_sort_function || post_sort_function(datap[heap[index]],count_out))
       {
-       itemsize=*(FILESORT_VARINT*)(datap[heap[index]]-FILESORT_VARSIZE);
+       itemsize=*(FILESORT_VARINT*)((char*)datap[heap[index]]-FILESORT_VARSIZE);
 
-       WriteFile(fd_out,datap[heap[index]]-FILESORT_VARSIZE,itemsize+FILESORT_VARSIZE);
-       count++;
+       WriteFileBuffered(fd_out,(char*)datap[heap[index]]-FILESORT_VARSIZE,itemsize+FILESORT_VARSIZE);
+       count_out++;
       }
 
-    if(ReadFile(fds[heap[index]],&itemsize,FILESORT_VARSIZE))
+    if(ReadFileBuffered(fds[heap[index]],&itemsize,FILESORT_VARSIZE))
       {
        heap[index]=heap[ndata];
        ndata--;
       }
     else
       {
-       *(FILESORT_VARINT*)(datap[heap[index]]-FILESORT_VARSIZE)=itemsize;
+       *(FILESORT_VARINT*)((char*)datap[heap[index]]-FILESORT_VARSIZE)=itemsize;
 
-       ReadFile(fds[heap[index]],datap[heap[index]],itemsize);
+       ReadFileBuffered(fds[heap[index]],datap[heap[index]],itemsize);
       }
 
     /* Bubble down the new value */
@@ -755,10 +861,10 @@ index_t filesort_vary(int fd_in,int fd_out,int (*compare)(const void*,const void
 
        newindex=2*index;
 
-       if(compare(datap[heap[newindex]],datap[heap[newindex+1]])>=0)
+       if(compare_function(datap[heap[newindex]],datap[heap[newindex+1]])>=0)
           newindex=newindex+1;
 
-       if(compare(datap[heap[index]],datap[heap[newindex]])<=0)
+       if(compare_function(datap[heap[index]],datap[heap[newindex]])<=0)
           break;
 
        temp=heap[newindex];
@@ -775,7 +881,7 @@ index_t filesort_vary(int fd_in,int fd_out,int (*compare)(const void*,const void
 
        newindex=2*index;
 
-       if(compare(datap[heap[index]],datap[heap[newindex]])<=0)
+       if(compare_function(datap[heap[index]],datap[heap[newindex]])<=0)
           ; /* break */
        else
          {
@@ -794,7 +900,7 @@ index_t filesort_vary(int fd_in,int fd_out,int (*compare)(const void*,const void
  if(fds)
    {
     for(i=0;i<nfiles;i++)
-       CloseFile(fds[i]);
+       CloseFileBuffered(fds[i]);
     free(fds);
    }
 
@@ -803,12 +909,16 @@ index_t filesort_vary(int fd_in,int fd_out,int (*compare)(const void*,const void
 
  for(i=0;i<option_filesort_threads;i++)
    {
-    free(threads[i].data);
+    log_free(threads[i].data);
 
-    free(threads[i].filename);
+    free(threads[i].data);
    }
 
- return(count);
+ free(threads);
+
+ free(filename);
+
+ return(count_out);
 }
 
 
@@ -822,20 +932,17 @@ index_t filesort_vary(int fd_in,int fd_out,int (*compare)(const void*,const void
 
 static void *filesort_fixed_heapsort_thread(thread_data *thread)
 {
- int fd,i;
+ size_t item;
 
  /* Sort the data pointers using a heap sort */
 
  filesort_heapsort(thread->datap,thread->n,thread->compare);
 
- /* Create a temporary file and write the result */
+ /* Write the result to the given temporary file */
 
- fd=OpenFileNew(thread->filename);
-
- for(i=0;i<thread->n;i++)
-    WriteFile(fd,thread->datap[i],thread->itemsize);
-
- CloseFile(fd);
+ if(thread->fd > 0)
+    for(item=0;item<thread->n;item++)
+       WriteFileBuffered(thread->fd,thread->datap[item],thread->itemsize);
 
 #if defined(USE_PTHREADS) && USE_PTHREADS
 
@@ -866,24 +973,21 @@ static void *filesort_fixed_heapsort_thread(thread_data *thread)
 
 static void *filesort_vary_heapsort_thread(thread_data *thread)
 {
- int fd,i;
+ size_t item;
 
  /* Sort the data pointers using a heap sort */
 
  filesort_heapsort(thread->datap,thread->n,thread->compare);
 
- /* Create a temporary file and write the result */
+ /* Write the result to the given temporary file */
 
- fd=OpenFileNew(thread->filename);
+ if(thread->fd > 0)
+    for(item=0;item<thread->n;item++)
+      {
+       FILESORT_VARINT itemsize=*(FILESORT_VARINT*)((char*)thread->datap[item]-FILESORT_VARSIZE);
 
- for(i=0;i<thread->n;i++)
-   {
-    FILESORT_VARINT itemsize=*(FILESORT_VARINT*)(thread->datap[i]-FILESORT_VARSIZE);
-
-    WriteFile(fd,thread->datap[i]-FILESORT_VARSIZE,itemsize+FILESORT_VARSIZE);
-   }
-
- CloseFile(fd);
+       WriteFileBuffered(thread->fd,(char*)thread->datap[item]-FILESORT_VARSIZE,itemsize+FILESORT_VARSIZE);
+      }
 
 #if defined(USE_PTHREADS) && USE_PTHREADS
 
@@ -915,20 +1019,20 @@ static void *filesort_vary_heapsort_thread(thread_data *thread)
 
   size_t nitems The number of items of data to sort.
 
-  int(*compare)(const void *, const void *) The comparison function (identical to qsort if the
-                                            data to be sorted was an array of things not pointers).
+  int (*compare_function)(const void*, const void*) The comparison function.  This is identical
+     to qsort if the data to be sorted is an array of things not pointers.
   ++++++++++++++++++++++++++++++++++++++*/
 
-void filesort_heapsort(void **datap,size_t nitems,int(*compare)(const void*, const void*))
+void filesort_heapsort(void **datap,size_t nitems,int(*compare_function)(const void*, const void*))
 {
  void **datap1=&datap[-1];
- int i;
+ size_t item;
 
  /* Fill the heap by pretending to insert the data that is already there */
 
- for(i=2;i<=nitems;i++)
+ for(item=2;item<=nitems;item++)
    {
-    int index=i;
+    size_t index=item;
 
     /* Bubble up the new value (upside-down, put largest at top) */
 
@@ -939,7 +1043,7 @@ void filesort_heapsort(void **datap,size_t nitems,int(*compare)(const void*, con
 
        newindex=index/2;
 
-       if(compare(datap1[index],datap1[newindex])<=0) /* reversed compared to filesort_fixed() above */
+       if(compare_function(datap1[index],datap1[newindex])<=0) /* reversed comparison to filesort_fixed() above */
           break;
 
        temp=datap1[index];
@@ -952,28 +1056,28 @@ void filesort_heapsort(void **datap,size_t nitems,int(*compare)(const void*, con
 
  /* Repeatedly pull out the root of the heap and swap with the bottom item */
 
- for(i=nitems;i>1;i--)
+ for(item=nitems;item>1;item--)
    {
-    int index=1;
+    size_t index=1;
     void *temp;
 
     temp=datap1[index];
-    datap1[index]=datap1[i];
-    datap1[i]=temp;
+    datap1[index]=datap1[item];
+    datap1[item]=temp;
 
     /* Bubble down the new value (upside-down, put largest at top) */
 
-    while((2*index)<(i-1))
+    while((2*index)<(item-1))
       {
        int newindex;
-       void *temp;
+       void **temp;
 
        newindex=2*index;
 
-       if(compare(datap1[newindex],datap1[newindex+1])<=0) /* reversed compared to filesort_fixed() above */
+       if(compare_function(datap1[newindex],datap1[newindex+1])<=0) /* reversed comparison to filesort_fixed() above */
           newindex=newindex+1;
 
-       if(compare(datap1[index],datap1[newindex])>=0) /* reversed compared to filesort_fixed() above */
+       if(compare_function(datap1[index],datap1[newindex])>=0) /* reversed comparison to filesort_fixed() above */
           break;
 
        temp=datap1[newindex];
@@ -983,14 +1087,14 @@ void filesort_heapsort(void **datap,size_t nitems,int(*compare)(const void*, con
        index=newindex;
       }
 
-    if((2*index)==(i-1))
+    if((2*index)==(item-1))
       {
        int newindex;
        void *temp;
 
        newindex=2*index;
 
-       if(compare(datap1[index],datap1[newindex])>=0) /* reversed compared to filesort_fixed() above */
+       if(compare_function(datap1[index],datap1[newindex])>=0) /* reversed comparison to filesort_fixed() above */
           ; /* break */
        else
          {
